import pygame
from pygame.locals import *
from Box2D.Box2D import *

#It is necessary to transform screen coordinates into box2d coordinates to improve stability
PIXELSPERMETER = 32.0 #32 pixels = 1 meter
METERSPERPIXEL = 1.0/32.0

#This is the main object all physics objects should derive from and it includes basic functionality all physics objects need
class PhysicsObject:

    def __init__(self, world, ID, x, y, screenheight, color, isStatic, hitpoints):
        self.world = world
        self.x = METERSPERPIXEL*x
        self.y = METERSPERPIXEL*(screenheight-y)
        self.screenheight = screenheight
        self.isStatic = isStatic
        self.maxHP = hitpoints
        self.hitpoints = hitpoints
        self.color = color
        self.ID = ID

    def draw(self, screen, screenheight):
        return

    def getIsStatic(self):
        return self.isStatic

    def getHP(self):
        return self.hitpoints

    def decreaseHP(self, x):
        if self.ID >= 0:
            self.hitpoints -= x

    def destroy(self):
        self.world.DestroyBody(self.body)

    def getBody(self):
        return self.body

    def getID(self):
        return self.ID

    def setAngle(self, angle):
        self.body.angle = angle

    def applyImpulse(self, x, y):
        pos = self.body.GetWorldPoint(b2Vec2(0,0))
        self.body.ApplyLinearImpulse(b2Vec2(x,y),pos,True)

#Represents a rectangle that is affected by physics
class PhysicsRectangle(PhysicsObject):
    
    def __init__(self, world, screenheight, x, y, width, height, color, ID=-1, angle=0, isStatic=False, hitpoints=100, density=1.0,friction=0.3):
        PhysicsObject.__init__(self, world, ID, x, y, screenheight, color, isStatic, hitpoints)

        #Convert screen coordinates into box2d coordinates
        self.width = METERSPERPIXEL*width
        self.height = METERSPERPIXEL*height
        
        if isStatic:
            self.body=world.CreateStaticBody(position=(self.x,self.y),angle=angle)
        else:
            self.body=world.CreateDynamicBody(position=(self.x,self.y),angle=angle)
        self.fixture = self.body.CreatePolygonFixture(box=(self.width/2,self.height/2), density=density, friction=friction)

        self.body.userData = self

    def draw(self, screen):
        #Transform the rectangles vertices using the info from box2d
        polygon = self.fixture.shape
        vertices=[(self.body.transform*v)*PIXELSPERMETER for v in polygon.vertices]
        vertices=[(v[0], self.screenheight-v[1]) for v in vertices]

        #Fade the box color based on how much health is left
        color = [0,0,0]
        i = 0
        for rgb in self.color:
            color[i] = rgb*(0.5+ (float(self.hitpoints)/self.maxHP) *0.5)
            if color[i] < 0:
                color[i] = 0
            i+=1
        
        pygame.draw.polygon(screen, color, vertices)

#Represents a circle that is affected by physics. Currently unused.
##class PhysicsCircle(PhysicsObject):
##    
##    def __init__(self, world, screenheight, x, y, radius, color, ID=-1, angle=0, isStatic=False, hitpoints=100, density=1.0,friction=0.3, angularDamping=0.01):
##        PhysicsObject.__init__(self, world, ID, x, y, screenheight, color, isStatic, hitpoints)
##
##        self.radius = METERSPERPIXEL*radius
##        
##        if isStatic:
##            self.body=world.CreateStaticBody(position=(self.x,self.y),angle=angle)
##        else:
##            self.body=world.CreateDynamicBody(position=(self.x,self.y),angle=angle)
##        self.body.angularDamping = angularDamping
##        self.fixture = self.body.CreateCircleFixture(radius=self.radius, density=density, friction=friction)
##        
##        self.body.userData = self
##
##    def draw(self, screen):
##        circle = self.fixture.shape
##        position=self.body.transform*circle.pos*PIXELSPERMETER
##        position=(position[0], self.screenheight-position[1])
##
##        color = [0,0,0]
##        i = 0
##        for rgb in self.color:
##            color[i] = rgb*(0.5+ (float(self.hitpoints)/self.maxHP) *0.5)
##            if color[i] < 0:
##                color[i] = 0
##            i+=1
##        
##        pygame.draw.circle(screen, color, [int(x) for x in position], int(circle.radius*PIXELSPERMETER))

#The physics object creator manages the physics objects and keeps track of them. It is also responsible for ensuring the HP of objects is
#decreased when the are in a big enough collision
class PhysicsObjectCreator(b2ContactListener):

    def __init__(self, screenheight):
        b2ContactListener.__init__(self)
        self.world = b2World(gravity=(0,-10),doSleep=True)
        self.world.contactListener = self
        self.screenheight = screenheight
        self.objectList = []
        self.objectsToDestroy = []
        self.isHP = True

    def toggleHPEnabled(self):
        self.isHP = not self.isHP

    def BeginContact(self, contact):
        pass
    def EndContact(self, contact):
        pass
    def PreSolve(self, contact, oldManifold):
        pass
    def PostSolve(self, contact, impulse):
        if (not self.isHP):
            return
        
        maxImpulse = 0
        for i in impulse.normalImpulses:
            if abs(i) > maxImpulse:
                maxImpulse = abs(i)

        #Sometimes there is a random huge impulse that should be ignored
        if maxImpulse > 1E10:
            maxImpulse = 1
        #If the impulse is big enough damage the objects
        if maxImpulse > 2:
            contact.fixtureA.body.userData.decreaseHP(maxImpulse)
            contact.fixtureB.body.userData.decreaseHP(maxImpulse)

    def update(self, timestep, velocityiter, constraintiter):
        self.world.Step(timestep, velocityiter, constraintiter)
        for obj in self.objectList:
            if obj.getHP() < 0:
                self.addObjectToDelete(obj) 

    def getIDCount(self):
        return 4

    #This method contains predefined types of physics objects that will be used in this game
    def createObjectByID(self, obid, x, y, isStatic=False, angle=0):
        if obid == -2:
            newobj = PhysicsRectangle(self.world, self.screenheight,x ,y ,32, 1300,(255,255,255),-2, 0,isStatic,0,0.1,0.1) #Walls
        if obid == -1:
            newobj = PhysicsRectangle(self.world, self.screenheight, x,y ,640,32,(255,255,255),-1, 0,isStatic) #Ground
        elif obid == 0:
            newobj = PhysicsRectangle(self.world, self.screenheight, x, y, 20, 60, (100,100,100), obid, angle, isStatic,200,3.0,0.6) #Stone
        elif obid == 1:
            newobj = PhysicsRectangle(self.world, self.screenheight, x, y, 14, 80, (162,113,38), obid, angle, isStatic,80) #Wood
        elif obid == 2:
            newobj = PhysicsRectangle(self.world, self.screenheight, x, y, 20, 20, (255,0,255), obid, angle, isStatic,100,2.0,0.7) #Enemy object
        elif obid == 3:
            newobj = PhysicsRectangle(self.world, self.screenheight, x, y, 23, 23, (255,0,0), obid, angle, isStatic,400,10.0,0.4) #Slingshot ammo

        self.objectList.append(newobj)
        return newobj


    #Cant just delete a physics object whenever because other calculations may depend on it.
    #Save the object to be deleted in a list and delete it later when it's safe.
    def addObjectToDelete(self, obj):
        if obj.getID() >= 0:
            self.objectsToDestroy.append(obj)

    #Actually delete the objects
    def deleteObjects(self):
            for delobj in self.objectsToDestroy:
                self.objectList.remove(delobj)
                delobj.destroy()
            self.objectsToDestroy = []

    #Get the object at a point or return false
    def getObjectAtPoint(self, x, y):
        x *= METERSPERPIXEL
        y = (self.screenheight - y) * METERSPERPIXEL

        for body in self.world.bodies:
            for fixture in body.fixtures:
                if (fixture.TestPoint((x, y))):
                    return body.userData
        return False

    def getObjectList(self):
        return self.objectList

    #Save the state of the physics to a file
    def saveCurrentState(self, levelNumber):
        f = open(str(levelNumber) + ".txt", "w")

        for obj in self.objectList:
            if not obj.getIsStatic():
                body = obj.getBody()
                ID = obj.getID()
                x = body.position.x*PIXELSPERMETER
                y = self.screenheight - body.position.y*PIXELSPERMETER
                angle = body.angle
                f.write(str(ID) + "," + str(x) + "," + str(y) + "," + str(angle) + "\n")
        
        f.close()

    #Load the state of physics from a file
    def loadState(self, levelNumber):
        try:
            f = open(str(levelNumber) + ".txt", "r")
        except IOError:
            return False

        for line in f:
            param = line.split(",")
            self.createObjectByID(int(param[0]), float(param[1]), float(param[2]), False, float(param[3]))

        f.close()

        #Prime the physics engine without HP because huge impulses that would break everything would result otherwise
        self.isHP = False
        for i in range(10000):
            self.update(1.0/60, 10, 10)
        self.isHP = True
        return True
