import pygame
from pygame.locals import *

#Screen is the base class which all other game screens inherit from. It defines the basic behaviour of a screen.
class Screen:

    def __init__(self, screenManager):
        self.manager = screenManager

    def update(self, timestep):
        return
    def draw(self, screen):
        return
    def keyinput(self, keypress, keyup, mousepress, mouseup):
        return
#This class is responsible for which screen is currently being displayed. The game is broken up into screens which contain the game's logic.
#Each screen is stored on a stack and the top screen is the one currently in use. This allows for switching between two screens without any
#complex saving logic ex. switching between the pause menu and the main game screen.
class ScreenManager:

    def __init__(self, targetFPS, width, height):
        #Initialize the screen stack
        self.screens = []
        self.running = False
        self.fps = targetFPS
        self.timestep = 1.0/targetFPS
        self.width = width
        self.height = height
        
        #Clock helps to keep the game running at a constant rate
        self.clock = clock=pygame.time.Clock()

    #This adds a screen to the top of the stack and makes it the current active screen.
    def addScreen(self, screen):
        self.screens.append(screen)
        
    #Removes a screen and sets the currently active screen to the screen before the one removed and if no screens are left the game is over.                
    def removeTopScreen(self):
        self.screens.pop()
        if len(self.screens) < 1:
            self.running = False

    #Calling run will run the top screen in the stack and effectively start the game
    def run(self):
        self.running = True

        #Create the main window
        screen=pygame.display.set_mode((self.width, self.height), 0, 32)
        pygame.display.set_caption('Block Attack')

        while self.running:
            #Ensure game is running at the correct FPS
            self.clock.tick(self.fps)

            keypressed = []
            keyreleased = []
            mousepressed = []
            mousereleased = []

            for event in pygame.event.get():
                #Exit the game loop if the user presses escape or closes the window
                if event.type == QUIT or (event.type==KEYDOWN and event.key==K_ESCAPE):
                    self.running = False
                elif event.type == KEYDOWN:
                    keypressed.append(event.key)
                elif event.type == KEYUP:
                    keyreleased.append(event.key)
                elif event.type == MOUSEBUTTONDOWN:
                    mousepressed.append(event.button)
                elif event.type == MOUSEBUTTONUP:
                    mousereleased.append(event.button)

            #Clear the screen
            screen.fill((0,0,0,0))
            
            topScreen = self.screens[-1]

            #Send input, update and then draw the top screen
            topScreen.keyinput(keypressed, keyreleased, mousepressed, mousereleased)
            topScreen.update(self.timestep)
            topScreen.draw(screen)

            #Display graphics to the screen
            pygame.display.flip()
